#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stddef.h>
#include <math.h>
#include <float.h>

#include "pds3.h"
#include "gds.h"
#include "bms.h"
#include "bds.h"
#include "grib.h"
#define VERSION "fix_grib_bds Wesley Ebisuzaki Version 1.1\n"


/*
 * fix_grib_bds:
 *
 *  fix bad bds_unused_bit flag
 *
 *                              Wesley Ebisuzaki
 *
 *
 */

/*
 * MSEEK = I/O buffer size for seek_grib
 */

#define MSEEK 1024
#define BUFF_ALLOC0	40000


#ifndef min
   #define min(a,b)  ((a) < (b) ? (a) : (b))
   #define max(a,b)  ((a) < (b) ? (b) : (a))
#endif
 
void main(int argc, char **argv) {

    unsigned char *buffer;
    int i, j, nx, ny;
    long int len_grib, pos = 0, nxny, buffer_size, count = 1;
    unsigned char *msg, *pds, *gds, *bms, *bds, *pointer;
    FILE *input;
    int n_mod = 0, modified;

    if (argc == 1) {
	fprintf(stderr, VERSION);
	exit(8);
    }

    if ((input = fopen(argv[1],"r+")) == NULL) {
        fprintf(stderr,"could not open file: %s\n", argv[1]);
        exit(7);
    }

    if ((buffer = (unsigned char *) malloc(BUFF_ALLOC0)) == NULL) {
	fprintf(stderr,"not enough memory\n");
    }
    buffer_size = BUFF_ALLOC0;


    for (;;) {
	msg = seek_grib(input, &pos, &len_grib, buffer, MSEEK);
	if (msg == NULL) break;

        /* read all whole grib record */
        if (len_grib + msg - buffer > buffer_size) {
            buffer_size = len_grib + msg - buffer + 1000;
            buffer = (unsigned char *) realloc((void *) buffer, buffer_size);
            if (buffer == NULL) {
                fprintf(stderr,"ran out of memory\n");
                exit(8);
            }
        }
        read_grib(input, pos, len_grib, buffer);
	modified = 0;

	/* parse grib message */

	msg = buffer;
        pds = (msg + 8);
        pointer = pds + PDSLEN(pds);

        if (HAS_GDS(pds)) {
            gds = pointer;
            pointer += GDS_LEN(gds);
        }
        else {
            gds = NULL;
        }

        if (HAS_BMS(pds)) {
            bms = pointer;
            pointer += BMS_LEN(bms);
        }
        else {
            bms = NULL;
        }

        bds = pointer;
        pointer += BDS_LEN(bds);

        /* end section - "7777" in ascii */
        if (pointer[0] != 0x37 || pointer[1] != 0x37 ||
            pointer[2] != 0x37 || pointer[3] != 0x37) {
            fprintf(stderr,"\n\n    missing end section\n");
            fprintf(stderr, "%2x %2x %2x %2x\n", pointer[0], pointer[1], 
		pointer[2], pointer[3]);
	    exit(8);
        }

	nx = GDS_LatLon_nx(gds);
	ny = GDS_LatLon_ny(gds);

	nxny = BDS_NValues(bds);
	if (bms != NULL) {
		nxny += missing_points(BMS_bitmap(bms),nx*ny);
	}
	if (nxny != nx*ny) {
	    i = (BDS_LEN(bds) - BDS_DataStart(bds))*8;
	    j = (nx*ny - missing_points(BMS_bitmap(bms),nx*ny))*BDS_NumBits(bds);
	    if (i > j && i - j < 16) {
		printf("fixable error in bds record %ld\n",count);
		bds[3] = (bds[3] & 0xf0) + (i - j);
                modified = 1;
	    }
	    else {
	        fprintf(stderr,"grib header %ld: two values of nxny %ld %d!\n",
                    count,nxny,nx*ny);
  	        fprintf(stderr,"   LEN %d DataStart %d UnusedBits %d #Bits %d\n",
		    BDS_LEN(bds), BDS_DataStart(bds),BDS_UnusedBits(bds),
		    BDS_NumBits(bds));
	    }
	}
	if (modified) {
	    n_mod++;
	    if (fseek(input,pos,SEEK_SET) == -1) {
		fprintf(stderr,"fatal error - i/o system\n");
		exit(12);
	    }
	    i = fwrite(msg, sizeof (unsigned char), len_grib, input);
	    if (i != len_grib) {
		fprintf(stderr,"fatal error - i/o system\n");
		exit(12);
	    }
	} 
        pos += len_grib;
        count++;
    }
    fclose(input);
    printf("number of records modified %d\n",n_mod);
}
